/**@file

// fwDIPSystemIntegrity.ctl
This library contains the system integrity functions of the fwDIPSystemIntegrity.

@par Creation Date
  26/11/2012

@par Modification History
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@author 
  EN/ICE/SIC (EN-ICE)
*/
//@{

// List of the fwDIPSystemIntegrity-DS dp checked
global dyn_string g_fwDIPSystemIntegrity_DIPAPIList;
// List of the fwDIPSystemIntegrity-DS thread Id checking the counter
global dyn_int g_fwDIPSystemIntegrity_ThreadId;

global int g_fwDIPSystemIntegrityCheckingDelay;
global int g_fwDIPSystemIntegrityWaitDelay;
global string g_fwDIPSystemIntegrity_startHierarchy;
global dyn_string g_fwDIPSystemIntegrity_publication_root;

/** Return the list of systemAlarm pattern. 
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@return list of systemAlarm pattern
*/
dyn_string fwDIPSystemIntegrity_systemIntegrityInfo()
{
  return makeDynString(DIP_pattern);
}

/** check if this a DIP API or DIP DNS config. 
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@param sType input, type of DP
@param sDIPConfig_1 output, sDIPConfig_1
@param sDIPConfig_2 output, sDIPConfig_2
@return true, if correct
*/
bool _fwDIPSystemIntegrity_accept(string sType, string &sDIPConfig_1, string &sDIPConfig_2)
{
  bool bReturn=false;
  dyn_string dsTemp;
  
  if(strpos(sType, "DNS") >=0)
  {
    dsTemp = strsplit(sType, "_");
    if(dynlen(dsTemp) == 3)
    {
      bReturn = dpExists("DIPConfig_"+dsTemp[2]);
      if(bReturn)
        bReturn = dpExists("DIPConfig_"+dsTemp[3]);
      sDIPConfig_1 = "DIPConfig_"+dsTemp[2];
      sDIPConfig_2 = "DIPConfig_"+dsTemp[3];
    }
    else
      bReturn = false;
  }
  else
  {
    bReturn = dpExists("DIPConfig_"+sType);
    sDIPConfig_1 = "DIPConfig_"+sType;
    sDIPConfig_2 = "";
  }
  return bReturn;
}

/** Get the list of fwDIPSystemIntegrity checked and enabled by the systemIntegrity. 
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@param dsResult output, list of enabled front-end systemIntegrity
*/
fwDIPSystemIntegrity_Initialize(dyn_string &dsResult)
{
  dyn_string dsList;
  int len, i;
  string dpToCheck;
  bool enabled;
  string sDIPConfig_1, sDIPConfig_2;

  dpGet(UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".config.data", dsList);
  fwDIPSystemIntegrity_DataCallback("", dsList);

  dsList = dpNames(c_unSystemAlarm_dpPattern+DIP_pattern+"*", c_unSystemAlarm_dpType);
  len = dynlen(dsList);
  for(i = 1; i<=len; i++) {
    dpToCheck = dpSubStr(dsList[i], DPSUB_DP);
    dpToCheck = substr(dpToCheck, strlen(c_unSystemAlarm_dpPattern+DIP_pattern),strlen(dpToCheck));
    if(_fwDIPSystemIntegrity_accept(dpToCheck, sDIPConfig_1, sDIPConfig_2)) {
      dpGet(dsList[i]+".enabled", enabled);
      if(enabled)
        dynAppend(dsResult, dpToCheck);
    }
  }
}

/** Callback on the configuration data
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@param sDpe1  input, dpe
@param dsConfigData input, configuration data
*/
fwDIPSystemIntegrity_DataCallback(string sDpe1, dyn_string dsConfigData)
{
  dyn_string dsTemp=dsConfigData;

  // check the size of the configuration data and add default value if some data are missing
  while(dynlen(dsTemp) <=2)
    dynAppend(dsTemp, "");
  strreplace(dsTemp[3], "//", "/");
  g_fwDIPSystemIntegrity_startHierarchy = strltrim(strrtrim(dsTemp[3],"/"),"/");
  g_fwDIPSystemIntegrityCheckingDelay = dsTemp[1];
  if(g_fwDIPSystemIntegrityCheckingDelay <= 0)
    g_fwDIPSystemIntegrityCheckingDelay = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYCheckingDelay;
  g_fwDIPSystemIntegrityWaitDelay = dsTemp[2];
  if(g_fwDIPSystemIntegrityWaitDelay <= 0)
    g_fwDIPSystemIntegrityWaitDelay = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYWaitDelay;
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity", "fwDIPSystemIntegrity_DataCallback", dsConfigData, g_fwDIPSystemIntegrityCheckingDelay, g_fwDIPSystemIntegrityWaitDelay, g_fwDIPSystemIntegrity_startHierarchy);
}

/** Callback on the systemIntegrity command.
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@param sDpe1  input, dpe
@param command input, type of command
@param sDpe2  input, dpe
@param parameters input, parameters of the command
*/
fwDIPSystemIntegrity_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
  dyn_string exceptionInfo;
  int i, len =dynlen(parameters);
  dyn_string dsResult;
   string sResult;
   string sMessage;
 
  switch(command) {
    case UN_SYSTEMINTEGRITY_ADD: // add the _UnSystemAlarm DP
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        for(i=1; i<=len; i++) {
          fwDIPSystemIntegrity_checking(parameters[i], true, true, exceptionInfo);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_DELETE: // delete the _UnSystemAlarm DP
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        for(i=1; i<=len; i++) {
          fwDIPSystemIntegrity_checking(parameters[i], false, false, exceptionInfo);
          // remove all _UnSystemAlarm dp from the alert list of applicationDP if it is in
          _unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern+DIP_pattern+parameters[i], false, exceptionInfo);
          // delete all the _UnSystemAlarm dps
          if(dpExists(c_unSystemAlarm_dpPattern+DIP_pattern+parameters[i]))
            dpDelete(c_unSystemAlarm_dpPattern+DIP_pattern+parameters[i]);
          _fwDIPSystemIntegrity_clean(parameters[i]);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_ENABLE: // enable the _UnSystemAlarm DP
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        for(i=1; i<=len; i++) {
          if(dpExists(c_unSystemAlarm_dpPattern+DIP_pattern+parameters[i]))
            fwDIPSystemIntegrity_checking(parameters[i], false, true, exceptionInfo);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_DISABLE: // disable the _UnSystemAlarm DP
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        for(i=1; i<=len; i++) {
          if(dpExists(c_unSystemAlarm_dpPattern+DIP_pattern+parameters[i]))
            fwDIPSystemIntegrity_checking(parameters[i], false, false, exceptionInfo);
        }
      }
      break;
    case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // give the list of _UnSystemAlarm DP and the state
      dpSet(UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
          UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", g_fwDIPSystemIntegrity_DIPAPIList);
      break;
    case UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC: // extended diagnostic
      dynAppend(dsResult, getCurrentTime());
      dynAppend(dsResult, "g_fwDIPSystemIntegrity_DIPAPIList:");
      len = dynlen(g_fwDIPSystemIntegrity_DIPAPIList);
      for(i=1;i<=len;i++) {
        sResult = g_fwDIPSystemIntegrity_DIPAPIList[i];
        dynAppend(dsResult, sResult);
      }
      dynAppend(dsResult, "g_fwDIPSystemIntegrity_ThreadId:");
      len = dynlen(g_fwDIPSystemIntegrity_ThreadId);
      for(i=1;i<=len;i++) {
        sResult = g_fwDIPSystemIntegrity_ThreadId[i];
        dynAppend(dsResult, sResult);
      }
      dynAppend(dsResult, "g_fwDIPSystemIntegrityCheckingDelay:" + g_fwDIPSystemIntegrityCheckingDelay);
      dynAppend(dsResult, "g_fwDIPSystemIntegrityWaitDelay:" + g_fwDIPSystemIntegrityWaitDelay);
      dynAppend(dsResult, "g_fwDIPSystemIntegrity_startHierarchy:" + g_fwDIPSystemIntegrity_startHierarchy);
      dynAppend(dsResult, "g_fwDIPSystemIntegrity_publication_root:");
      len = dynlen(g_fwDIPSystemIntegrity_publication_root);
      for(i=1;i<=len;i++) {
        sResult = g_fwDIPSystemIntegrity_publication_root[i];
        dynAppend(dsResult, "DIP pub: "+sResult);
      }
      dpSet(UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
          UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", dsResult);
      DebugTN(dsResult);
      break;
    case UN_SYSTEMINTEGRITY_DEBUG:
      if(dynlen(parameters) >= 2)
      {
        g_b32DebugLevel = parameters[1];
        g_sDebugFilter = parameters[2];
      }
      else
      {
        g_b32DebugLevel = 0;
        g_sDebugFilter = "";
      }
      break;
    case UN_SYSTEMINTEGRITY_VERSION:
      sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("fwDIPSystemIntegrity"));
      dpSet(UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
            UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", makeDynString(getCurrentTime(),sMessage));
      break;
    default:
      break;
  }

  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity", "fwDIPSystemIntegrity_HandleCommand", command, parameters);
  if(dynlen(exceptionInfo) > 0) {
    if(isFunctionDefined("unMessageText_sendException")) {
      unMessageText_sendException("*", "*", "fwDIPSystemIntegrity_HandleCommand", "user", "*", exceptionInfo);
    }
// handle any error uin case the send message failed
    if(dynlen(exceptionInfo) > 0) {
      DebugN(getCurrentTime(), exceptionInfo);
    }
  }
}

/** This function register/de-register the callback funtion of fwDIPSystemIntegrity manager and creates the _unSystemAlarm_fwDIPSystemIntegrity dp 
with the alarm config but it does not delete it.

  
@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sDp  input, data point name
@param bCreate  input, true create the dp and the alarm config if it does not exist, enable it
@param bRegister  input, true do a dpConnect, false do a dpDisconnect
@param exceptionInfo output, exception are returned here
*/
fwDIPSystemIntegrity_checking(string sDp, bool bCreate, bool bRegister, dyn_string &exceptionInfo)
{
  string sDpToChek, dp, dpPlc, plcHostname, description;
  int res, thId, pos, len,i;
  bool bError = false;

  string alertPanel,alertHelp;
  bool configExists, isActive;
  int alertConfigType;
  dyn_string alertTexts,alertClasses,summaryDpeList,alertPanelParameters;
  dyn_float alertLimits;
  dyn_string dsTemp;
  string sDIPConfig_1, sDIPConfig_2;
  
// remove the system name
  res = strpos(sDp, ":");
  if(res >= 0)
    dp = substr(sDp, res+1, strlen(sDp));
  else
    dp = sDp;
  // x == system integrity on DIP API
  // DNS_y_z == DIP DNS, y primary, z secondary
  if(!_fwDIPSystemIntegrity_accept(dp, sDIPConfig_1, sDIPConfig_2))
    return;
  sDpToChek = dp;

  dp = c_unSystemAlarm_dpPattern+DIP_pattern+dp;

  if(bCreate) {
// get the plc hostname and the driver number
    plcHostname = unGenericDpFunctions_getAlias(sDpToChek);
// create all the _UnSystemIntegrity Dp and its alarm config if it is not existing
    if(!dpExists(dp)) {
      if(strpos(sDpToChek, "DNS_") >=0) 
      {
        dsTemp = strsplit(sDpToChek, "_");
        while(dynlen(dsTemp)  <= 3)
          dynAppend(dsTemp, "?");
        description = "DIP API "+dsTemp[2]+" on Primary DNS - API "+dsTemp[3]+" on Secondary DNS";
      }
      else
        description = "DIP API "+sDpToChek;
      unSystemIntegrity_createSystemAlarm(sDpToChek, DIP_pattern, description, exceptionInfo);
    }
  }

  if(bRegister) {
// connect to the callback function if not already done
    pos = _unSystemIntegrity_isInList(g_fwDIPSystemIntegrity_DIPAPIList, dp);
    if(pos <= 0) {
  // add it in the list, because one of the callback function needs it.        
      pos = _unSystemIntegrity_setList(g_fwDIPSystemIntegrity_DIPAPIList, dp, true);

// start the thread for checking the counter
      thId = startThread(UN_SYSTEM_INTEGRITY_FWDIPSYSTEMINTEGRITY_check, sDpToChek);
      g_fwDIPSystemIntegrity_ThreadId[pos] = thId;
      if(thId<0)
        bError = true;

      if(bError) {
        pos = _unSystemIntegrity_isInList(g_fwDIPSystemIntegrity_DIPAPIList, dp);
        if(pos > 0) {
    // kill the threads if they were started
          res = stopThread(g_fwDIPSystemIntegrity_ThreadId[pos]);
// and then remove it from the list
          pos = _unSystemIntegrity_setList(g_fwDIPSystemIntegrity_DIPAPIList, dp, false);
//DebugN("rmv", res);
          dynRemove(g_fwDIPSystemIntegrity_ThreadId, pos);
        }

        fwException_raise(exceptionInfo, "ERROR", 
              "fwDIPSystemIntegrity_checking():"+getCatStr("unSystemIntegrity", "CANNOT_START_THREAD") +dp,"");
      }
      else {
//DebugN("Enable", g_fwDIPSystemIntegrity_ThreadId, "end");
      // set the enable to true and activate the alarm if it is not activated of all the _UnSystemIntegrity DP.
        dpSet(dp+".enabled", true);
        unAlarmConfig_mask(dp+".alarm", false, exceptionInfo);
      }
    }
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity", "fwDIPSystemIntegrity_checking register", sDp,
                                 g_fwDIPSystemIntegrity_ThreadId, g_fwDIPSystemIntegrity_ThreadId);
  }
  else {
// disconnect the callback function
    pos = _unSystemIntegrity_isInList(g_fwDIPSystemIntegrity_DIPAPIList, dp);
    if(pos > 0) {
// kill the thread
      res = stopThread(g_fwDIPSystemIntegrity_ThreadId[pos]);
      if(res<0)
        bError = true;

      if(bError) {
        fwException_raise(exceptionInfo, "ERROR", 
              "fwDIPSystemIntegrity_checking():"+getCatStr("unSystemIntegrity", "CANNOT_STOP_THREAD") +dp,"");
      }
      else {
      // remove from list
        pos = _unSystemIntegrity_setList(g_fwDIPSystemIntegrity_DIPAPIList, dp, false);
//DebugN("rmv", res);
        dynRemove(g_fwDIPSystemIntegrity_ThreadId, pos);
      // set the enable to false and de-activate the alarm and acknowledge it if necessary of all the _UnSystemIntegrity DP.
        dpSet(dp+".enabled", false, dp+".alarm", c_unSystemIntegrity_no_alarm_value);
        unAlarmConfig_mask(dp+".alarm", true, exceptionInfo);
      }
    }
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity", "fwDIPSystemIntegrity_checking deregister", sDp, 
                                 g_fwDIPSystemIntegrity_DIPAPIList, g_fwDIPSystemIntegrity_ThreadId);
  }
}

/** This function checks the state of DIP API. 
A _UnSystemAlarm is set if the counter is not updated periodically.

@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sParam  input, data to check
*/
fwDIPSystemIntegrityCheck(string sParam)
{
  if(strpos(sParam, "DNS_") >=0)
    fwDIPSystemIntegrityCheck_DIPDNS(sParam);
  else
    fwDIPSystemIntegrityCheck_DIP(sParam);
}

/** This function checks the state of DIP API. 
A _UnSystemAlarm is set if the counter is not updated periodically.

@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sAPINum  input, DIP API num
*/
fwDIPSystemIntegrityCheck_DIP(string sAPINum)
{
  int waitingTime, oldAlarmValue = -1, oldTimingAlarm = -1;
  int oldValue = -1, newValue, alarmValue, iTime;
  bool bEnabled, bTemp;
  string sModPlc;
  int res, iPlcNumber;
  string sDIP_DP;
  int waitDelay;
  string sRootPub;
  int iPubTime;
  string sDIPConfig_1, sDIPConfig_2;
  
  // create DIP publication & subscription DP
  sDIP_DP = FWDIPDPE_DPTYPE+"_"+sAPINum;
  sRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_DP);
  if(!dpExists(sDIP_DP))
    dpCreate(sDIP_DP, FWDIPDPE_DPTYPE);
  _fwDIPSystemIntegrity_addRemove(sDIP_DP, true);
  bTemp = _fwDIPSystemIntegrity_accept(sAPINum, sDIPConfig_1, sDIPConfig_2);
  fwDIPSystemIntegrity_publication(sDIPConfig_1, sDIP_DP, sRootPub, DIP_pattern+sAPINum, true);
  while(true) {
  // if 0 set it to g_fwDIPSystemIntegrityCheckingDelay
    waitingTime = g_fwDIPSystemIntegrityCheckingDelay;
    if(waitingTime <= 0)
      waitingTime = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYCheckingDelay;
    waitDelay = g_fwDIPSystemIntegrityWaitDelay;
    if(waitDelay <= 0)
      waitDelay = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYWaitDelay;
  // wait
    delay(waitingTime);
    if(dpExists(sDIP_DP))
    {
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        iPubTime = getCurrentTime();
      // running DIP manager: create publication, create subscription
        _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_DP, sRootPub, DIP_pattern+sAPINum, true);
        delay(waitDelay);
        oldValue = iPubTime;
        dpSet(sDIP_DP+".iPublication", oldValue);
        delay(waitDelay);
        dpGet(sDIP_DP+".iSubscription", newValue);
      // no alarm if publication equal subscription 
      // alarm 10 if publication not equal subscription
        if(newValue == oldValue) {
          alarmValue = c_unSystemIntegrity_no_alarm_value;
        }
        else {
          alarmValue = c_unSystemIntegrity_alarm_value_level1;
        }
      // delete publication, subscription
        _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_DP, sRootPub, DIP_pattern+sAPINum, false);
      }
    }
    else // 50: internal error
      alarmValue = 5*c_unSystemIntegrity_alarm_value_level1;

    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sAPINum, "fwDIPSystemIntegrityCheck_DIP check", sAPINum, 
                                 waitingTime, waitDelay, oldValue, newValue, oldAlarmValue, alarmValue, sDIP_DP,
                                 sRootPub, iPubTime);
    
    if(oldAlarmValue != alarmValue) {
      dpSet(c_unSystemAlarm_dpPattern+DIP_pattern+sAPINum+".alarm", alarmValue);
    }

    oldAlarmValue = alarmValue;
  }
}

/** This function checks the state of DIP DNS. 
A _UnSystemAlarm is set if the counter is not updated periodically.

@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sAPINum  input, data point name
*/
fwDIPSystemIntegrityCheck_DIPDNS(string sAPINum)
{
  int waitingTime, oldAlarmValue = -1, oldTimingAlarm = -1;
  int primaryOldValue = -1, primaryNewValue, secondaryOldValue = -1, secondaryNewValue, alarmValue, iTime;
  bool bEnabled, bTemp;
  string sModPlc;
  int res, iPlcNumber;
  string sDIP_primary_DNS;
  string sDIP_secondary_DNS;
  int waitDelay, iPubTime;
  string sPrimaryRootPub, sSecondaryRootPub;
  string sDIPConfig_1, sDIPConfig_2;
  
  //primary DNS  create DIP publication & subscription DP
  _fwDIPSystemIntegrity_getDIPDNS_DP(sAPINum, true, sDIP_primary_DNS);
  _fwDIPSystemIntegrity_addRemove(sDIP_primary_DNS, true);
  sPrimaryRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_primary_DNS);
  if(!dpExists(sDIP_primary_DNS))
    dpCreate(sDIP_primary_DNS, FWDIPDPE_DPTYPE);
  //secondary DNS  create DIP publication & subscription DP
   _fwDIPSystemIntegrity_getDIPDNS_DP(sAPINum, false, sDIP_secondary_DNS);
  _fwDIPSystemIntegrity_addRemove(sDIP_secondary_DNS, true);
  sSecondaryRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_secondary_DNS);
  if(!dpExists(sDIP_secondary_DNS))
    dpCreate(sDIP_secondary_DNS, FWDIPDPE_DPTYPE);
  bTemp = _fwDIPSystemIntegrity_accept(sAPINum, sDIPConfig_1, sDIPConfig_2);
  fwDIPSystemIntegrity_publication(sDIPConfig_1, sDIP_primary_DNS, sPrimaryRootPub, sAPINum, true);
  fwDIPSystemIntegrity_publication(sDIPConfig_2, sDIP_secondary_DNS, sSecondaryRootPub, sAPINum, true);
  while(true) {
  // if 0 set it to g_fwDIPSystemIntegrityCheckingDelay
    waitingTime = g_fwDIPSystemIntegrityCheckingDelay;
    if(waitingTime <= 0)
      waitingTime = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYCheckingDelay;
    waitDelay = g_fwDIPSystemIntegrityWaitDelay;
    if(waitDelay <= 0)
      waitingTime = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYWaitDelay;
  // wait
    delay(waitingTime);
    if((dpExists(sDIP_primary_DNS)) && (dpExists(sDIP_secondary_DNS)))
    {
      synchronized(g_fwDIPSystemIntegrity_ThreadId)
      {
        iPubTime = getCurrentTime();
      // DNS checking
      // primary DNS: create publication, create subscription
      // primary DNS: set publication value, get subscription value
        _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_primary_DNS, sPrimaryRootPub, sAPINum, true);
        delay(waitDelay);
        primaryOldValue = iPubTime;
        dpSet(sDIP_primary_DNS+".iPublication", primaryOldValue);
        delay(waitDelay);
        dpGet(sDIP_primary_DNS+".iSubscription", primaryNewValue);
  
      // secondary DNS: create publication, create subscription
      // secondary DNS: set publication value, get subscription value
        _fwDIPSystemIntegrity_subscription(sDIPConfig_2, sDIP_secondary_DNS, sSecondaryRootPub, sAPINum, true);
        delay(waitDelay);
        secondaryOldValue = iPubTime;
        dpSet(sDIP_secondary_DNS+".iPublication", secondaryOldValue);
        delay(waitDelay);
        dpGet(sDIP_secondary_DNS+".iSubscription", secondaryNewValue);
  
      // no alarm if on primary DNS publication equal subscription and on secondary publication not equal to subscription
      // alarm 20 if on secondary DNS publication equal subscription and on primary publication not equal to subscription
      // alarm 30 if on secondary DNS publication equal subscription and on primary publication equal to subscription
      // alarm 40 if on secondary DNS publication not equal subscription and on primary publication not equal to subscription
         if((primaryNewValue == primaryOldValue) && (secondaryNewValue != secondaryOldValue))
         {
           alarmValue = c_unSystemIntegrity_no_alarm_value;
         }
        else if((primaryNewValue != primaryOldValue) && (secondaryNewValue == secondaryOldValue))
        {
          alarmValue = 2*c_unSystemIntegrity_alarm_value_level1;
        }
        else if((primaryNewValue == primaryOldValue) && (secondaryNewValue == secondaryOldValue))
        {
          alarmValue = 3*c_unSystemIntegrity_alarm_value_level1;
        }
         else
           alarmValue = 4*c_unSystemIntegrity_alarm_value_level1;
   
      // primary DNS: delete publication, subscription
        _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_primary_DNS, sPrimaryRootPub, sAPINum, false);
      // secondary DNS: delete publication, subscription
        _fwDIPSystemIntegrity_subscription(sDIPConfig_2, sDIP_secondary_DNS, sSecondaryRootPub, sAPINum, false);
      }
    }
    else
    { // 50: internal error
      alarmValue = 5*c_unSystemIntegrity_alarm_value_level1;
    }
    
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sAPINum, "fwDIPSystemIntegrityCheck_DIPDNS check", sAPINum, 
                                 waitingTime, waitDelay, primaryNewValue, primaryOldValue, secondaryNewValue, secondaryOldValue, oldAlarmValue, alarmValue, 
                                 sDIP_primary_DNS, sDIP_secondary_DNS, sPrimaryRootPub, sSecondaryRootPub, iPubTime);
    
    if(oldAlarmValue != alarmValue) {
      dpSet(c_unSystemAlarm_dpPattern+DIP_pattern+sAPINum+".alarm", alarmValue);
    }

    oldAlarmValue = alarmValue;
  }
}

/** clean the DIP publication and subscription.

@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sAPINum  input, data point name
*/
_fwDIPSystemIntegrity_clean(string sAPINum)
{
  string sDIP_publication, sRootPub, sDIPConfig_1, sDIPConfig_2;
  bool bTemp;
  
  if(strpos(sAPINum, "DNS_") >=0)
  {
    bTemp = _fwDIPSystemIntegrity_accept(sAPINum, sDIPConfig_1, sDIPConfig_2);
    // primary DNS: delete publication, subscription
    _fwDIPSystemIntegrity_getDIPDNS_DP(sAPINum, true, sDIP_publication);
    _fwDIPSystemIntegrity_addRemove(sDIP_publication, false);
    sRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_publication);
    fwDIPSystemIntegrity_publication(sDIPConfig_1, sDIP_publication, sRootPub, sAPINum, false);
    _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_publication, sRootPub, sAPINum, false);
    // primary DNS: delete DPs
    if(dpExists(sDIP_publication))
      dpDelete(sDIP_publication);
    // secondary DNS: delete publication
    _fwDIPSystemIntegrity_getDIPDNS_DP(sAPINum, false, sDIP_publication);
    _fwDIPSystemIntegrity_addRemove(sDIP_publication, false);
    sRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_publication);
    _fwDIPSystemIntegrity_subscription(sDIPConfig_2, sDIP_publication, sRootPub, sAPINum, false);
    fwDIPSystemIntegrity_publication(sDIPConfig_2, sDIP_publication, sRootPub, sAPINum, false);
    // secondary DNS: delete DPs
    if(dpExists(sDIP_publication))
      dpDelete(sDIP_publication);
  }
  else
  {
    bTemp = _fwDIPSystemIntegrity_accept(sAPINum, sDIPConfig_1, sDIPConfig_2);
    // delete publication
    sDIP_publication = FWDIPDPE_DPTYPE+"_"+sAPINum;
    _fwDIPSystemIntegrity_addRemove(sDIP_publication, false);
    sRootPub =_fwDIPSystemIntegrity_getPubName(sDIP_publication);
    fwDIPSystemIntegrity_publication(sDIPConfig_1, sDIP_publication, sRootPub, DIP_pattern+sAPINum, false);
    _fwDIPSystemIntegrity_subscription(sDIPConfig_1, sDIP_publication, sRootPub, DIP_pattern+sAPINum, false);
    // delete DPs
    if(dpExists(sDIP_publication))
      dpDelete(sDIP_publication);
    // delete DIP publication & subscription DP
  }
}

/** add the DIP publication in the internal list
@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sPubName  input, DIP publication
@param bAdd  input, true=add/false=remove
*/
_fwDIPSystemIntegrity_addRemove(string sPubName, bool bAdd) 
{
  int iPos;
  string sInternalPub = _fwDIPSystemIntegrity_getPubName(sPubName);
  synchronized(g_fwDIPSystemIntegrity_publication_root)
  {
    if(bAdd)
      dynAppend(g_fwDIPSystemIntegrity_publication_root, sInternalPub);
    else
    {
      iPos = dynContains(g_fwDIPSystemIntegrity_publication_root, sInternalPub);
      dynRemove(g_fwDIPSystemIntegrity_publication_root, iPos);
    }
    dynUnique(g_fwDIPSystemIntegrity_publication_root);
  }
}

/** add the DIP subscription
@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sDIPConfigDp  input, DIP config DP
@param sDP  input, DIP publication/subscription DP
@param sRootName  input, DIP root publication
@param sPubName  input, DIP publication
@param bAdd  input, true=add/false=remove
*/
_fwDIPSystemIntegrity_subscription(string sDIPConfigDp, string sDP, string sRootName, string sPubName, bool bCreate) 
{
  int waitDelay;
  string sName = sRootName+"/"+sPubName;
  dyn_string exceptionInfo;
  
  waitDelay = g_fwDIPSystemIntegrityWaitDelay;
  if(waitDelay <= 0)
    waitDelay = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYWaitDelay;
  if(bCreate)
  {
    // create sub
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sDIPConfigDp, "_fwDIPSystemIntegrity_subscription", "creating sub", sDIPConfigDp, sDP+".iSubscription", sName);
    fwDIP_subscribe(sDP+".iSubscription",sDIPConfigDp, sName, "", exceptionInfo, true);
  }
  else
  {
    // delete sub
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sDIPConfigDp, "_fwDIPSystemIntegrity_subscription", "deleting sub", sDP+".iSubscription", sName);
    fwDIP_unsubscribe(sDP+".iSubscription", exceptionInfo);
  }
}

/** add the DIP publication
@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sDIPConfigDp  input, DIP config DP
@param sDP  input, DIP publication/subscription DP
@param sRootName  input, DIP root publication
@param sPubName  input, DIP publication
@param bAdd  input, true=add/false=remove
*/
fwDIPSystemIntegrity_publication(string sDIPConfigDp, string sDP, string sRootName, string sPubName, bool bCreate) 
{
  int waitDelay;
  string sName = sRootName+"/"+sPubName;
  dyn_string exceptionInfo;
  
  waitDelay = g_fwDIPSystemIntegrityWaitDelay;
  if(waitDelay <= 0)
    waitDelay = c_fwSystemIntegrity_defaultFWDIPSYSTEMINTEGRITYWaitDelay;
  if(bCreate)
  {
    // create pub
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sDIPConfigDp, "_fwDIPSystemIntegrity_subscription", "creating pub", sDIPConfigDp, sDP+".iPublication", sName);
    fwDIP_publishPrimitive(sName, sDP+".iPublication", 0, sDIPConfigDp, exceptionInfo, true);
  }
  else
  {
    // delete pub
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "fwDIPSystemIntegrity "+sDIPConfigDp, "_fwDIPSystemIntegrity_subscription", "deleting pub", sDP+".iPublication", sName);
    fwDIP_unpublish(sName, exceptionInfo);
  }
}

/** get the DIP publication name.
@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sPubName  input, name
@return the publication name
*/
string _fwDIPSystemIntegrity_getPubName(string sPubName)
{
  string sInternalPub = "dip/"+g_fwDIPSystemIntegrity_startHierarchy+"/"+strrtrim(getSystemName(), ":")+"/"+getSystemId()+"/"+sPubName;
  strreplace(sInternalPub, "//", "/"); 
  return sInternalPub;
}
/** get the DIP DP publication and subscription for DNS check.

@par Constraints
  None

@par Usage
  Internal

@par PVSS managers
  CTRL

@param sDIPDNS  input, DNS
@param bPrimary  input, primary DNS
@param sDIP_DP  output, publication DP
*/
_fwDIPSystemIntegrity_getDIPDNS_DP(string sDIPDNS, bool bPrimary, string &sDIP_DP)
{
  dyn_string dsTemp=strsplit(sDIPDNS, "_");
  
  if(dynlen(dsTemp) == 3)
  {
    if(bPrimary)
    {
      sDIP_DP = FWDIPDPE_DPTYPE+"_primary_"+sDIPDNS;
    }
    else
    {
      sDIP_DP = FWDIPDPE_DPTYPE+"_secondary_"+sDIPDNS;
    }
  }
  else
  {
    sDIP_DP = "";
  }
}

//@}
